/*
 * This source file contains C code of an example HW/SW interface for the 
 * Xilinx XUP board.  
 * 
 * Redistributions of any form whatsoever must retain and/or include the
 * following acknowledgment, notices and disclaimer:
 *
 * This product includes software developed by Carnegie Mellon University.
 *
 * Copyright (c) 2006 by J. C. Hoe, Carnegie Mellon University
 *
 * You may not use the name "Carnegie Mellon University" or derivations
 * thereof to endorse or promote products derived from this software.
 *
 * If you modify the software you must place a notice on or within any
 * modified version provided or made available to any third party stating
 * that you have modified the software.  The notice shall include at least
 * your name, address, phone number, email address and the date and purpose
 * of the modification.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" WITHOUT ANY WARRANTY OF ANY KIND, EITHER
 * EXPRESS, IMPLIED OR STATUTORY, INCLUDING BUT NOT LIMITED TO ANY WARRANTY
 * THAT THE SOFTWARE WILL CONFORM TO SPECIFICATIONS OR BE ERROR-FREE AND ANY
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
 * TITLE, OR NON-INFRINGEMENT.  IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY
 * BE LIABLE FOR ANY DAMAGES, INCLUDING BUT NOT LIMITED TO DIRECT, INDIRECT,
 * SPECIAL OR CONSEQUENTIAL DAMAGES, ARISING OUT OF, RESULTING FROM, OR IN
 * ANY WAY CONNECTED WITH THIS SOFTWARE (WHETHER OR NOT BASED UPON WARRANTY,
 * CONTRACT, TORT OR OTHERWISE).
 */


#include <stdio.h>
#include <assert.h>
#include <xparameters.h>
#include <xio.h>

#define OCM0BASE ((int*)XPAR_DOCM_CNTLR_BASEADDR)
#define OCMMEMSIZE (4*1024) // in bytes

volatile int *OcmInBase = OCM0BASE + (0*OCMMEMSIZE/sizeof(int));
volatile int *OcmOutBase = OCM0BASE + (1*OCMMEMSIZE/sizeof(int));
volatile int *OcmCntrlBase = OCM0BASE + (2*OCMMEMSIZE/sizeof(int));
int *OcmMemBase = OCM0BASE + (3*OCMMEMSIZE/sizeof(int));

#define IBUFHD (0)
#define IBUFTL (8)
#define OBUFHD (16)
#define OBUFTL (24)
#define SFTRST (56)

#define SYNC {asm volatile ("sync"::);} 

// cacheblock aligned buffer
int buff1[4*OCMMEMSIZE/sizeof(int)] __attribute__ ((aligned(32)));
int buff2[4*OCMMEMSIZE/sizeof(int)] __attribute__ ((aligned(32)));

int __ibufhead=0;
int __ibuftail=(-1%(OCMMEMSIZE/sizeof(int)));
int __obufhead=0;
int __obuftail=0;

void init() {
  static int cache=0;
  printf(" --- reset HW --- \n");

  // enable caching, assume 512MByte
  if (!cache) {
    XCache_EnableICache(0xcc000000);
    XCache_EnableDCache(0xa0000000);
    cache=1;
  }

  SYNC;
  OcmCntrlBase[SFTRST]=-1;
  SYNC;
  OcmCntrlBase[IBUFHD]=0;
  OcmCntrlBase[OBUFTL]=0;

  __ibufhead=0;
  __ibuftail=(-1%(OCMMEMSIZE/sizeof(int)));
  __obufhead=0;
  __obuftail=0;

  SYNC
  OcmCntrlBase[SFTRST]=0;
  SYNC;
}

void dumpState(char *str) {
  unsigned int temp;
  printf("---%s---\n", str);
  // read hardware registers
  printf("ibufhd=%x --- ", OcmCntrlBase[IBUFHD]); // head pointer (next to pop) for SW->HW
  printf("ibuftl=%x --- ", OcmCntrlBase[IBUFTL]); // tail pointer (next to push) for SW->HW
  printf("obufhd=%x --- ", OcmCntrlBase[OBUFHD]); // head pointer (next to pop) for HW->SW
  printf("obuftl=%x\n", OcmCntrlBase[OBUFTL]);    // tail pointer (next to push) for HW->SW
  printf("sftrst=%x\n", OcmCntrlBase[SFTRST]);

  // read software "cached" copies
  printf("softptr ibufhd=%x --- ", __ibufhead);
  printf("softptr ibuftl=%x --- ", __ibuftail);
  printf("softptr obufhd=%x --- ", __obufhead);
  printf("softptr obuftl=%x\n", __obuftail);
}

/* Pushing from buffer "num" integer words to the OCM HW.
 * This is a blocking call.
 */
void push(int *buffer, int num) {
  int index=0;
  int released=0;
  int numx=num;

  while(num--) {
    while (__ibufhead==__ibuftail) {
      if (!released) {
	SYNC;
	OcmCntrlBase[IBUFHD]=__ibufhead;
	released=1;
      }
      SYNC;
      __ibuftail=OcmCntrlBase[IBUFTL];
      __ibuftail--;
      __ibuftail%=(OCMMEMSIZE/sizeof(int));
    }
    released=0;

    OcmInBase[__ibufhead]=buffer[index++];
    __ibufhead++;
    __ibufhead%=(OCMMEMSIZE/sizeof(int));
  }
  SYNC;
  OcmCntrlBase[IBUFHD]=__ibufhead;
}

/* Receiving into buffer "num" integer words from the OCM HW.
 * This is a blocking call.
 */
void pop(int *buffer, int num) {
  int index=0;
  int released=0;

  while(num--) {
    while (__obufhead==__obuftail) {
      if (!released) {
	SYNC;
	OcmCntrlBase[OBUFTL]=__obuftail;
	released=1;
      }
      SYNC;
      __obufhead=OcmCntrlBase[OBUFHD];
    }
    released=0;

    buffer[index++]=OcmOutBase[__obuftail];
    __obuftail++;
    __obuftail%=(OCMMEMSIZE/sizeof(int));
  }
  SYNC;
  OcmCntrlBase[OBUFTL]=__obuftail;
}

/* example code to exercise and check the interface */
void testefifo() {
  int iter=0, size=0, dataseed=10;

  printf(" --- test EFIFO --- \n");

  for(iter=0;iter<10000;iter++) {
    int idx;

    // initialize buff1 with data
    size=(size+1)%2000;
    idx=size;

    while(idx--) {
      buff1[idx]=dataseed++;
    }

    // push buff1 and pop buff2, demo HW is just does a loopback
    push(buff1,size);
    pop(buff2,size);

    // check returned buff2 against original buff1
    idx=size;
    while(idx--) {
      if (!(buff1[idx]==buff2[idx])) {
	int k;
	for(k=-2;k<2;k++) {
	  printf("%d: buff1[%x]==buff2[%x]\n",idx-k, buff1[idx-k],buff2[idx-k]);
	}
	for(k=-2;k<2;k++) {
	  int where=(__ibufhead-size+(idx-k))%(OCMMEMSIZE/sizeof(int));
	  assert(__ibufhead==__obuftail);
	  printf("%d %d %d %d\n",__ibufhead,idx,k,size);
	  printf("ocmi[%d]=%x::ocmo[%d]=%x\n",
		 where,
		 OcmInBase[where],
		 where,
		 OcmOutBase[where]);
	}
	dumpState("Bailing out\n");
	printf("size=%d\n",size);
      }
      assert(buff1[idx]==buff2[idx]);
    }

    if (!(size%100)) printf(".");
  }
  putchar('\n');
}

extern int test_mmm();

int main (void) {
  int iter=0;

  init();
  
  test_mmm();  // test the reference implementation of matrix multiplication

  while(1) {
    printf(" --- Iteration %d --- \n", iter++);
    testefifo();
  }

  return(0);
}

