/*
 * mmm.c - blocked matrix-matrix multiply
 *
 * Redistributions of any form whatsoever must retain and/or include the
 * following acknowledgment, notices and disclaimer:
 *
 * This product includes software developed by Carnegie Mellon University.
 *
 * Copyright (c) 2006 by Roland E. Wunderlich at Carnegie Mellon University
 *
 * You may not use the name "Carnegie Mellon University" or derivations
 * thereof to endorse or promote products derived from this software.
 *
 * If you modify the software you must place a notice on or within any
 * modified version provided or made available to any third party stating
 * that you have modified the software.  The notice shall include at least
 * your name, address, phone number, email address and the date and purpose
 * of the modification.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" WITHOUT ANY WARRANTY OF ANY KIND, EITHER
 * EXPRESS, IMPLIED OR STATUTORY, INCLUDING BUT NOT LIMITED TO ANY WARRANTY
 * THAT THE SOFTWARE WILL CONFORM TO SPECIFICATIONS OR BE ERROR-FREE AND ANY
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
 * TITLE, OR NON-INFRINGEMENT.  IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY
 * BE LIABLE FOR ANY DAMAGES, INCLUDING BUT NOT LIMITED TO DIRECT, INDIRECT,
 * SPECIAL OR CONSEQUENTIAL DAMAGES, ARISING OUT OF, RESULTING FROM, OR IN
 * ANY WAY CONNECTED WITH THIS SOFTWARE (WHETHER OR NOT BASED UPON WARRANTY,
 * CONTRACT, TORT OR OTHERWISE).
 *
 */

/*
 *
 * This source file is distributed "as is" in the hope that it will be
 * useful.  The tool set comes with no warranty, and no author or
 * distributor accepts any responsibility for the consequences of its
 * use.
 *
 * Everyone is granted permission to copy, modify and redistribute
 * this tool set under the following conditions:
 *
 *    This source code is distributed for non-commercial use only.
 *    Please contact the maintainer for restrictions applying to
 *    commercial use.
 *
 *    Permission is granted to anyone to make or distribute copies
 *    of this source code, either as received or modified, in any
 *    medium, provided that all copyright notices, permission and
 *    nonwarranty notices are preserved, and that the distributor
 *    grants the recipient permission for further redistribution as
 *    permitted by this document.
 *
 *    Permission is granted to distribute this file in compiled
 *    or executable form under the same conditions that apply for
 *    source code, provided that either:
 *
 *    A. it is accompanied by the corresponding machine-readable
 *       source code,
 *    B. it is accompanied by a written offer, with no time limit,
 *       to give anyone a machine-readable copy of the corresponding
 *       source code in return for reimbursement of the cost of
 *       distribution.  This written offer must permit verbatim
 *       duplication by anyone, or
 *    C. it is distributed by someone who received only the
 *       executable form, and is accompanied by a copy of the
 *       written offer of source code that they received concurrently.
 *
 * In other words, you are welcome to use, share and improve this
 * source file.  You are forbidden to forbid anyone else to use, share
 * and improve what you give them.
 *
 */


// uncomment the next line to disable assertion checking
// #define NDEBUG

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>

#include "timer.h"

// The 32-bit Number type is a complex value
// we use 16-bit fixed point format for real (upper) and imaginary (lower)
typedef signed long Number;
#define WIDTH (16)
#define MASK (0xffff)
#define PACK(r,i) (((r)<<WIDTH)|((i)&MASK))
#define UNPACKR(r) ((r)>>WIDTH)
#define UNPACKI(i) (((i)<<WIDTH)>>WIDTH)

#define OVERFLOW(f) (!((((f)&0xffff8000)==0)||(((f)&0xffff8000)==0xffff8000)))

// maximum MMM size (for allocating buffers)
#define MMSIZE 1024

static inline Number complexMult(Number A, Number B) {
        // Our interpretation is
	//   B and the return value use the same 2's complement fixed-point format
	//   A uses 2's complement fixed-point format with 1-bit sign, 1-bit integer, 14-bit fraction
	signed long Ar=UNPACKR(A);
	signed long Ai=UNPACKI(A);
	signed long Br=UNPACKR(B);
	signed long Bi=UNPACKI(B);
	Number Cr, Ci;

	Cr=Ar*Br-Ai*Bi;
	Ci=Ar*Bi+Ai*Br;	

	Cr=Cr>>(WIDTH-2);
	Ci=Ci>>(WIDTH-2);

	assert(!OVERFLOW(Cr));
	assert(!OVERFLOW(Ci));

	return PACK(Cr,Ci);
}

static inline Number complexAdd(Number A, Number B) {
	// A, B and the return value use the same 2's complement fixed-point format
	signed long Ar=UNPACKR(A);
	signed long Ai=UNPACKI(A);
	signed long Br=UNPACKR(B);
	signed long Bi=UNPACKI(B);
	Number Cr, Ci;

	Cr=Ar+Br;
	Ci=Ai+Bi;

	assert(!OVERFLOW(Cr));
	assert(!OVERFLOW(Ci));

	return PACK(Cr,Ci);
}

static inline Number complexRandom() {
	Number Cr, Ci;
	unsigned long x=0;

	// no random in EDK stdlib
	//Cr=(Number)(random()%1024)-512;
	//Ci=(Number)(random()%1024)-512;

	// initialize to whatever
	Cr=(Number)((x+=751)%1024)-512;
	Ci=(Number)((x+=751)%1024)-512;

	return PACK(Cr,Ci);
}

// -----------------------------------------------------------------------------

/* Multiplies a [NB x NB] block of two square, row-major order matricies (A, B), 
	where the size of matrix A, B, and C is [N X N]. */

void mmmKernel(Number* A, Number* B, Number* C, int N, int NB) {	
        int i, j, k;
	for (j = 0; j < NB; j++)
		for (i = 0; i < NB; i++)
			for (k = 0; k < NB; k++)
				C[i * N + j] = complexAdd( C[i * N + j], complexMult(A[i * N + k], B[k * N + j]) );
}



// -----------------------------------------------------------------------------

/* Multiplies two square, row-major order matricies (A, B), where the size of
    matrix A and matrix B is [N X N], where 0 < N <= 16,384. The resulting 
    matrix is allocated using calloc, and should be freed by the caller. */

Number BUFC1[MMSIZE*MMSIZE];

Number* mmm(Number* A, Number* B, int N) {
	// verify all arguments, sizes of A and B are not checked
	assert(A != NULL);
	assert(B != NULL);
	assert((N > 0) && (N <= 0x4000));

	// allocate a zero-initialized block of memory for the result matrix
	// Number* C = (Number*) calloc(N * N, sizeof(Number));
	Number *C=BUFC1; // no malloc in EDK
	assert(C != NULL);
	
	// perform the matrix-matrix multiply on the whole matrix at once
	mmmKernel(A, B, C, N, N);
	
	return C;
}



// -----------------------------------------------------------------------------

/* Multiplies two square, row-major order matricies (A, B), where the size of
    matrix A and matrix B is [N X N], where 0 < N <= 16,384. The matrix 
    multiplication is blocked, where each block is [NB X NB], with NB > 0 and 
    N % NB == 0. The resulting matrix is allocated using calloc, and should be 
    freed by the caller. */
    
Number BUFC2[MMSIZE*MMSIZE];

Number* mmmBlocked(Number* A, Number* B, int N, int NB) {
	// verify all arguments, sizes of A and B are not checked
	assert(A != NULL);
	assert(B != NULL);
	assert((N > 0) && (N <= 0x4000));
	assert((NB > 0) && (N % NB == 0));

	// allocate a zero-initialized block of memory for the result matrix
	// Number* C = (Number*) calloc(N * N, sizeof(Number));
	Number *C=BUFC2;  // no malloc in EDK
	assert(C != NULL);

	{
		int j, i, k;
		// perform the matrix-matrix multiply one block of C at a time
		for (j = 0; j < N; j += NB)
			for (i = 0; i < N; i += NB)
				for (k = 0; k < N; k += NB)
					mmmKernel(&(A[i * N + k]), &(B[k * N + j]), &(C[i * N + j]), N, NB);
	}
	return C;
}



// -----------------------------------------------------------------------------

Number BUFA[MMSIZE*MMSIZE*2];

int test_mmm() {


	int N = 256; 
	int NB = 16; // note: N and NB must be evenly divisible)

	long long start, finish;
	
        printf(" --- test blocked MMM %d x %d--- \n", N, N);

	// allocate enouch space for two [N X N] matricies
	assert (sizeof(Number)==4);
	assert (sizeof(short)==2);
	//Number* A = (Number*) malloc(N * N * sizeof(Number) * 2);
	Number *A=BUFA; // no malloc in EDK
	assert(N<=MMSIZE);
	assert(A != NULL);

	{
		int x;
		// initialize matricies with random numbers
		for (x = 0; x < N * N * 2; x++) {
			A[x] = complexRandom(); 
		}
	}

	Number* B = &(A[N * N]);
	
	// perform a regular and a blocked matrix-matric multiply
	
	Number* referenceC;
	Number* blockedC;
	
	start=ts_get_ll();
	referenceC = mmm(A, B, N);
	finish=ts_get_ll();
	printf("Reference %d by %d mmm, unblocked: %d us\n", N, N, ts_elapse_us(start,finish));
	
	start=ts_get_ll();
	blockedC = mmmBlocked(A, B, N, NB);
	finish=ts_get_ll();
	printf("Reference %d by %d mmm, blocked by %d: %d us\n", N, N, NB, ts_elapse_us(start,finish));

	{
		int x;	
		// verify the two matrix-matrix multiplies are the same
		for (x = 0; x < N * N; x++)
			assert(blockedC[x] == referenceC[x]);
	}
		
	
	// exit
	// free(A);
	// free(referenceC);
	// free(blockedC);
	return 0;
}


